import json
import os
import threading
import time

from mqtool.kit import Kafka_consumer, Kafka_producer
from mqtool.kit.push_metrics import IO_metrics
from mqtool.kit.tornado_server import Tornado_http
from mqtool.model.algo_load import load_dir
from mqtool.utils import *

algo_list = load_dir('./algorithm')
if len(algo_list) == 0:
    logger.info('未加载到算法模型，程序退出！')
    exit()
algo = algo_list[0]
## 统计算子IO指标
input = 0
output = 0
input_befor = 0
output_befor = 0
time_differ_list = []
data_type = ''


def start():
    return {"status": "algorithm service success"}


def api(data):
    """
    处理data参数，算法运行
    :param data: dict字段类型
    :return:
    """
    if data_type == "kafka":
        if 'picture' in data:
            fdfsKey = data['picture']
            img = readfile(str(fdfsKey))
            data['img'] = img
    elif data_type == "http":
        if 'data' in data:
            img = readfrombase64(data['data'])
            del data['data']
            data['img'] = img
        if 'structure' in data:
            data['structure'] = data['structure'].replace('\"', '')
    return algo.run(data)


def cron_push():
    """
    定时push metrics
    :return:
    """
    global input, output, input_befor, output_befor,time_differ_list
    push_addr = os.getenv("PUSHTOGATEWAY", None)
    if push_addr is None:
        logger.info("None pushgateway address")
    if push_addr is not None:
        info = "pushgateway address:" + push_addr
        logger.info(info)
        while True:
            time.sleep(60)
            IO_metrics(push_addr,input,output,input_befor,output_befor,time_differ_list)
            input_befor,output_befor= input,output
            time_differ_list = []



def push_task():
    """
    开启新线程推送指标
    :return:
    """
    try:
        th = threading.Thread(target=cron_push)
        th.start()
    except:
        logger.exception("push metrics error")

def aistart():
    try:
        # push_task()
        data_type = os.getenv("DATA_TYPE", default="http")
        if data_type == "kafka":
            bootstrap_server = os.getenv("BOOTSTRAP_SERVER")
            kafkatopic = os.getenv("TOPIC")
            storage_url = os.getenv("STORAGE_URL", default="http://localhost:8888/")
            consumer = Kafka_consumer(bootstrap_servers=bootstrap_server, kafkatopic=kafkatopic,
                                      groupid=algo.__name__)
            data = consumer.read_df()
            sinktopic = os.getenv("SINKTOPIC")
            if sinktopic is not None:
                producer = Kafka_producer(bootstrap_servers=bootstrap_server, kafkatopic=sinktopic,
                                          is_json=True)
            for message in data:
                try:
                    input += 1
                    logger.info('receive data')
                    mes = json.loads(message.decode())
                    time_befor=time.time()
                    res = api(mes)
                    time_after = time.time()
                    time_differ_list.append(time_after-time_befor)
                    if res is not None:
                        if isinstance(res, list):
                            for item in res:
                                if sinktopic is not None:
                                    producer.send_df(item)
                                http_post(storage_url, item)
                                output +=1
                        else:
                            if sinktopic is not None:
                                producer.send_df(res)
                            http_post(storage_url, res)
                            output += 1
                except:
                    logger.exception("deal data error")

        elif data_type == "http":
            tornado_server = Tornado_http()
            # 添加路由
            tornado_server.add_handler("/", start)
            # 数据为json格式
            tornado_server.add_handler("/predict", api, json_type=True, get=False)
            # 开启服务
            tornado_server.start_http_server()
    except:
        logger.exception("system error")



from ctypes import *

import random



def sample(probs):
    s = sum(probs)
    probs = [a / s for a in probs]
    r = random.uniform(0, 1)
    for i in range(len(probs)):
        r = r - probs[i]
        if r <= 0:
            return i
    return len(probs) - 1


def c_array(ctype, values):
    arr = (ctype * len(values))()
    arr[:] = values
    return arr


class BOX(Structure):
    _fields_ = [("x", c_float),
                ("y", c_float),
                ("w", c_float),
                ("h", c_float)]


class DETECTION(Structure):
    _fields_ = [("bbox", BOX),
                ("classes", c_int),
                ("prob", POINTER(c_float)),
                ("mask", POINTER(c_float)),
                ("objectness", c_float),
                ("sort_class", c_int)]


class IMAGE(Structure):
    _fields_ = [("w", c_int),
                ("h", c_int),
                ("c", c_int),
                ("data", POINTER(c_float))]


class METADATA(Structure):
    _fields_ = [("classes", c_int),
                ("names", POINTER(c_char_p))]


lib = CDLL('/app/algorithm/libdarknet.so', RTLD_GLOBAL)
lib.network_width.argtypes = [c_void_p]
lib.network_width.restype = c_int
lib.network_height.argtypes = [c_void_p]
lib.network_height.restype = c_int

predict = lib.network_predict
predict.argtypes = [c_void_p, POINTER(c_float)]
predict.restype = POINTER(c_float)

set_gpu = lib.cuda_set_device
set_gpu.argtypes = [c_int]

make_image = lib.make_image
make_image.argtypes = [c_int, c_int, c_int]
make_image.restype = IMAGE

get_network_boxes = lib.get_network_boxes
get_network_boxes.argtypes = [c_void_p, c_int, c_int, c_float, c_float, POINTER(c_int), c_int, POINTER(c_int)]
get_network_boxes.restype = POINTER(DETECTION)

make_network_boxes = lib.make_network_boxes
make_network_boxes.argtypes = [c_void_p]
make_network_boxes.restype = POINTER(DETECTION)

free_detections = lib.free_detections
free_detections.argtypes = [POINTER(DETECTION), c_int]

free_ptrs = lib.free_ptrs
free_ptrs.argtypes = [POINTER(c_void_p), c_int]

network_predict = lib.network_predict
network_predict.argtypes = [c_void_p, POINTER(c_float)]

reset_rnn = lib.reset_rnn
reset_rnn.argtypes = [c_void_p]

load_net = lib.load_network
load_net.argtypes = [c_char_p, c_char_p, c_int]
load_net.restype = c_void_p

do_nms_obj = lib.do_nms_obj
do_nms_obj.argtypes = [POINTER(DETECTION), c_int, c_int, c_float]

do_nms_sort = lib.do_nms_sort
do_nms_sort.argtypes = [POINTER(DETECTION), c_int, c_int, c_float]

free_image = lib.free_image
free_image.argtypes = [IMAGE]

letterbox_image = lib.letterbox_image
letterbox_image.argtypes = [IMAGE, c_int, c_int]
letterbox_image.restype = IMAGE

load_meta = lib.get_metadata
lib.get_metadata.argtypes = [c_char_p]
lib.get_metadata.restype = METADATA

load_image = lib.load_image_color
load_image.argtypes = [c_char_p, c_int, c_int]
load_image.restype = IMAGE

rgbgr_image = lib.rgbgr_image
rgbgr_image.argtypes = [IMAGE]

predict_image = lib.network_predict_image
predict_image.argtypes = [c_void_p, IMAGE]
predict_image.restype = POINTER(c_float)

ndarray_image = lib.ndarray_to_image
ndarray_image.argtypes = [POINTER(c_ubyte), POINTER(c_long), POINTER(c_long)]
ndarray_image.restype = IMAGE


def nparray_to_image(img):
    data = img.ctypes.data_as(POINTER(c_ubyte))
    image = ndarray_image(data, img.ctypes.shape, img.ctypes.strides)
    return image


def classify(net, meta, im):
    out = predict_image(net, im)
    res = []
    for i in range(meta.classes):
        res.append((meta.names[i], out[i]))
    res = sorted(res, key=lambda x: -x[1])
    return res


def detect(net, meta, im, thresh=.5, hier_thresh=.5, nms=.45):
    num = c_int(0)
    pnum = pointer(num)
    predict_image(net, im)
    dets = get_network_boxes(net, im.w, im.h, thresh, hier_thresh, None, 0, pnum)
    num = pnum[0]
    if (nms): do_nms_obj(dets, num, meta.classes, nms);

    res = []
    for j in range(num):
        for i in range(meta.classes):
            if dets[j].prob[i] > 0:
                b = dets[j].bbox
                res.append((meta.names[i], dets[j].prob[i], (b.x, b.y, b.w, b.h)))
    res = sorted(res, key=lambda x: -x[1])
    free_image(im)
    free_detections(dets, num)
    return res




import importlib
import os

# ignore_files = ['model_service.py','demo.py','__init__.py']
ignore_files = ['model_service.py','__init__.py']

def load_dir(path):
    algo_list = []
    dirs = os.listdir(path)
    for file in dirs:
        if file.endswith('.py') and not file in ignore_files:
            mol = load_source(file[0:len(file) - 3], path + '/' + file)
            # algo_map[file.name] =mol
            if hasattr(mol,'run'):
                algo_list.append(mol)
    return algo_list


def load_source(name, path):
    return importlib.machinery.SourceFileLoader(name, path).load_module()



from prometheus_client import CollectorRegistry, Gauge, push_to_gateway



def IO_metrics(push_addr,input,output,input_befor,output_befor,time_differ_list):

    with open("/etc/hostname","r") as f:
        pod_name=f.read()
    registry = CollectorRegistry()
    if len(time_differ_list)==0:
        process_time=0
    else:
        process_time=sum(time_differ_list)/len(time_differ_list)
    input_rate= (input-input_befor)/60
    output_rate=(output-output_befor)/60
    h = Gauge('algorithmIO', '算子IO指标采集', ["io_type"],
              registry=registry)
    h.labels("input").set(input)
    h.labels("output").set(output)
    h.labels("input_rate").set(input_rate)
    h.labels("output_rate").set(output_rate)
    h.labels("process_time").set(process_time)

    push_to_gateway(push_addr, job=pod_name.strip(), registry=h)
    print("push metrics:{input:%s,output:%s,input_rate:%s,output_rate:%s,process_time:%s}"%(input,output,input_rate,output_rate,process_time))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File  : tornado_server.py
# @Desc  :  tornado  service
import json

import tornado.web
import tornado.ioloop

from mqtool.utils import logger


class Tornado_http:

    def __init__(self, port=8888):

        self.port = port
        self.handlers = []
        # self.executor = ThreadPoolExecutor(max_workers=1)

    def set_port(self, port):
        self.port = port

    # 添加路由
    def add_handler(self, handler_addr, re_fun, *args, json_type=False, post=True, get=True):
        """
        :param handler_addr: 路由地址
        :param re_fun: 视图函数
        :param args: 路由参数
        :return:
          """

        class handlername(tornado.web.RequestHandler):
            def get(self):
                if not get:
                    raise tornado.web.HTTPError(500)
                argsl = []
                for arg in args:
                    argsl.append(self.get_query_argument(arg))
                try:
                    res = re_fun(*argsl)
                    logger.info("res:" + str(res))
                    if res is None:
                        self.write({"algorithm":"nodata"})
                    elif isinstance(res, list):
                        self.set_header("Content-Type", "application/json; charset=UTF-8")
                        self.write(json.dumps(res,ensure_ascii=False))
                    elif isinstance(res, dict):
                        self.set_header("Content-Type", "application/json; charset=UTF-8")
                        self.write(json.dumps(res, ensure_ascii=False))
                    else:
                        self.write(res)
                except:
                    logger.exception("algorithm error")
                    self.write({"algorithm":"error"})

            def post(self):
                if not post:
                    raise tornado.web.HTTPError(500)
                try:
                    if json_type:
                        data = json.loads(self.request.body.decode('utf-8'))
                        res = re_fun(data)
                    else:
                        argsl = []
                        for arg in args:
                            argsl.append(self.get_argument(arg))
                        res = re_fun(*argsl)
                    logger.info("res:" + str(res))
                    if res is None:
                        self.write({"algorithm":"nodata"})
                    elif isinstance(res, list):
                        self.set_header("Content-Type", "application/json; charset=UTF-8")
                        self.write(json.dumps(res, ensure_ascii=False))
                    elif isinstance(res, dict):
                        self.set_header("Content-Type", "application/json; charset=UTF-8")
                        self.write(json.dumps(res, ensure_ascii=False))
                    else:
                        self.write(res)
                except:
                    logger.exception("algorithm error")
                    self.write({"algorithm":"error"})
        self.handlers.append((handler_addr, handlername))

    # 开启 tornado 服务
    def start_http_server(self):
        app = tornado.web.Application(self.handlers)
        app.listen(port=self.port, address="0.0.0.0")
        # 启动web程序，开始监听端口的连接
        # self.executor.submit(tornado.ioloop.IOLoop.current().start)
        logger.info("service start")
        tornado.ioloop.IOLoop.current().start()



from kafka import KafkaProducer
from kafka import KafkaConsumer
from kafka.errors import KafkaError
import json


class Kafka_producer():
    """
    生产模块：根据不同的key，区分消息
    """

    def __init__(self, bootstrap_servers, kafkatopic, key=None, is_json=False):
        self.bootstrap_servers = bootstrap_servers
        self.kafkatopic = kafkatopic
        self.key = key
        self.is_json = is_json
        if is_json:
            self.producer = KafkaProducer(bootstrap_servers=self.bootstrap_servers,
                                          value_serializer=lambda v: json.dumps(v,ensure_ascii=False).encode())
        else:
            self.producer = KafkaProducer(bootstrap_servers=self.bootstrap_servers)

    def send_df(self, message, topic=None):
        try:
            if topic is None:
                topic = self.kafkatopic
            producer = self.producer
            if self.is_json:
                producer.send(topic, message, key=self.key)
            else:
                producer.send(topic, key=self.key, value=message.encode())  # 注意转换
            producer.flush()
        except KafkaError as e:
            print(e)


class Kafka_consumer():

    def __init__(self, bootstrap_servers, kafkatopic, groupid=None, key=None, auto_offset_reset='latest'):
        self.bootstrap_servers = bootstrap_servers
        self.kafkatopic = kafkatopic
        self.groupid = groupid
        self.key = key
        self.consumer = KafkaConsumer(self.kafkatopic, group_id=self.groupid,
                                      auto_offset_reset=auto_offset_reset, bootstrap_servers=self.bootstrap_servers,
                                      enable_auto_commit=False)

    def assign(self):
        assignment = []
        count = 0
        while len(assignment) == 0 and count < 100:
            self.consumer.poll(100)
            assignment = self.consumer.assignment()
            count += 1
        return assignment

    def read_df(self, seek_end=True):
        if seek_end:
            assignment = self.assign()
            if len(assignment) > 0:
                self.consumer.seek_to_end(*assignment)
        try:
            for message in self.consumer:
                self.consumer.commit()
                yield message.value  # 返回迭代器
        except Exception as e:
            print(e)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 BussanQ

import sys
from os.path import dirname, join

try: # for pip >= 10
    from pip._internal.req import parse_requirements
except ImportError: # for pip <= 9.0.3
    from pip.req import parse_requirements

from setuptools import (
    find_packages,
    setup,
)


requirements = [str(ir.req) for ir in parse_requirements("requirements.txt", session=False)]


setup(
    name='mqtool',
    version='1.0.0',
    description='bq tool',
    packages=find_packages(exclude=[]),
    author='BussanQ',
    license='BussanQ',
    package_data={'': ['*.*']},
    install_requires=requirements,
    zip_safe=False,
    classifiers=[
        'Programming Language :: Python',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3',
    ],
)



import json
import logging
from uuid import uuid1
import requests
import base64
import numpy as np
import cv2
from fdfs_client.client import *


def http_post(url, json_data):
    json_data = json.dumps(json_data)
    with requests.session() as s:
        s.keep_alive = False
        headers = {'Content-Type': 'application/json'}
        response = s.post(url=url, data=json_data, headers=headers)
        json_res = response.json()
        logger.info("http_post result:"+str(json_res))
    return json_res

def readfile(imgId):
    tracker_path = get_tracker_conf('/etc/fdfs/client.conf')
    client = Fdfs_client(tracker_path)
    ret_read = client.download_to_buffer(imgId.encode())
    nparr = np.fromstring(ret_read['Content'], np.uint8)
    img_decode = cv2.imdecode(nparr, cv2.IMREAD_COLOR)
    # img = cv2.cvtColor(img_decode, cv2.COLOR_BGR2RGB)
    return img_decode


def readfrombase64(imgstr):
    if 'base64,' in imgstr:
        i = imgstr.index('base64,') + 7
        imgstr = imgstr[i:]
    imgData = base64.decodebytes(imgstr.encode())
    nparr = np.fromstring(imgData, np.uint8)
    img_decode = cv2.imdecode(nparr, cv2.IMREAD_COLOR)
    return img_decode


def uploadfile(thumbnail):
    tracker_path = get_tracker_conf('/etc/fdfs/client.conf')
    client = Fdfs_client(tracker_path)
    img_encode = cv2.imencode('.jpg', thumbnail)[1]
    data_encode = np.array(img_encode)
    str_encode = data_encode.tostring()
    subimg_key = client.upload_by_buffer(str_encode)
    thumbnail_id = subimg_key['Remote file_id']
    return thumbnail_id

def init_log():
    logger = logging.getLogger()
    logger.setLevel(level=logging.INFO)
    if len(logger.handlers) == 0:
        # handler = logging.FileHandler("log.txt")
        handler = logging.StreamHandler()
        handler.setLevel(logging.INFO)
        # pathname
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(filename)s:%(lineno)d[%(funcName)s] - %(message)s')
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    return logger

logger = init_log()


# 生成8位短uuid
uuidChars = ("a", "b", "c", "d", "e", "f",
             "g", "h", "i", "j", "k", "l", "m", "n", "o", "p", "q", "r", "s",
             "t", "u", "v", "w", "x", "y", "z", "0", "1", "2", "3", "4", "5",
             "6", "7", "8", "9", "A", "B", "C", "D", "E", "F", "G", "H", "I",
             "J", "K", "L", "M", "N", "O", "P", "Q", "R", "S", "T", "U", "V",
             "W", "X", "Y", "Z")


def short_uuid():
    uuid = str(uuid1()).replace('-', '')
    result = ''
    for i in range(0, 8):
        sub = uuid[i * 4: i * 4 + 4]
        x = int(sub, 16)
        result += uuidChars[x % 0x3E]
    return result


def get_short_uuids(length):
    uuid_list = [short_uuid() for i in range(length)]
    return uuid_list


def isEmpty(obj):
    """
    对象是否为空
    :param obj:
    :return:
    """
    return True if (obj is None or len(obj) == 0) else False


# # 暂定
# def get_db(user, password, ip, instance='orcl', port=1521):
#     engine_str = 'oracle://%s:%s@%s:%s/%s' % (user, password, ip, port, instance)
#     engine = create_engine(engine_str, encoding='utf-8')
#     return engine
#
#
# # 暂定
# def get_mysql(user, password, ip, db, port=3306):
#     engine_str = f'mysql+pymysql://{user}:{password}@{ip}:{port}/{db}'
#     engine = create_engine(engine_str, encoding='utf-8')
#     return engine
#
#
# # 暂定
# def get_pg(user="postgres", password="", ip="10.2.111.1", db="mydb", port=5432):
#     engine_str = f'postgresql+psycopg2://{user}:{password}@{ip}:{port}/{db}'
#     engine = create_engine(engine_str, encoding='utf-8')
#     return engine
